from io import StringIO
from pathlib import Path
import xml.etree.ElementTree as ET
from collections import defaultdict
from jinja2 import Environment, BaseLoader


class TestSuite:
    """TestSuite: container for all testsuite test cases. Normally, this is only one."""

    def __init__(self, xml):
        self.meta = xml.attrib
        self.testcases = defaultdict(list)
        for tc in xml.findall("./testcase"):
            tcobject = TestCase(tc)
            self.testcases[tcobject.classname].append(tcobject)
        self.testcases = dict(self.testcases)

    def __repr__(self):
        return "testsuite errors={errors} failures={failures} skipped={skipped} tests={tests} time={time} timestamp={timestamp}".format(
            **self.meta
        )


class TestCase:
    def __init__(self, xml):
        for k, v in xml.attrib.items():
            setattr(self, k, v)
        self.properties = {}
        for prop in xml.findall("./properties/property"):
            self.properties[prop.attrib["name"]] = prop.attrib["value"]
        # ---------------------------------------------------------------
        # skipped
        skipped = xml.find("./skipped")
        if skipped is not None:
            self.skipped = skipped.attrib
        else:
            self.skipped = None
        # ---------------------------------------------------------------
        # failed
        failure = xml.find("./failure")
        if failure is not None:
            self.failure = failure.attrib
        else:
            self.failure = None

    def __repr__(self):
        s = f"{self.classname}::{self.name}"
        if self.skipped:
            s += " (skipped)"
        if self.failure:
            s += " (failed)"
        return s


def parse_xml(filepath):
    """parse JUnit  XML file and return a tuple of `TestSuite` instances"""
    fpath = Path(filepath)
    assert fpath.exists()
    tree = ET.parse(filepath)
    root = tree.getroot()
    # get all testsuites
    return tuple((TestSuite(ts) for ts in root.findall("./testsuite")))


MASTER_TPL = """
MYSTRAN TESTING
===============

{% for testsuite in testsuites %}
Test Suite "{{ testsuite.meta["name"] }}"
----------------------------------

{% for ini_name, testcases in testsuite.testcases.items() %}
### {{ ini_name }}
{% for tc in testcases %}
#### {{ tc }}
{% if tc.skipped %}**SKIPPED**: {{tc.skipped['message']}} {% endif %}
{% if tc.failure %}**FAILED**: \n{{tc.failure['message']}} {% endif %}
{% endfor %}

{% endfor %}{# end of test cases #}
{% endfor %}{# end of test suites #}
"""


def xml2md(xmlfpath):
    tss = parse_xml(xmlfpath)
    rtemplate = Environment(loader=BaseLoader).from_string(MASTER_TPL)
    return rtemplate.render({"testsuites": tss})



import pandas as pd


def subset(df, **levels):
    """easily subset a dataframe"""
    levelnames = df.index.names
    unknown = set(levels.keys()) - set(levelnames)
    if unknown:
        levels = {k: v for k, v in levels.items() if k in levelnames}
    # validate levels
    filters = []
    for levelname in levelnames:
        if levelname not in levels or levels[levelname] is None:
            filter = slice(None)
        else:
            filter = levels[levelname]
            if isinstance(filter, str):
                filters.append([filter])
                continue
            try:
                filter = list(filter)
            except TypeError:
                filter = [filter]
        filters.append(filter)
    return df.loc[tuple(filters), :]



# -*- coding: utf-8 -*-

"""Console script for mystran_validation.

MYSTRAN Binary is found with the following scheme:

    * from `--mystran-bin` passed option
    * from "MYSTRAN_BIN" environment variable
    * from /usr/bin/mystran
"""
import configparser
import glob
import logging
import os
import shlex
import shutil
import subprocess
import sys
import webbrowser
from pathlib import Path
import pkg_resources

import click
import pytest

from mystran_validation import get_conf, get_profile, init_config


def find(
    path,
    extensions,
    name="*",
    break_on_first=True,
    return_first=True,
):
    """return a list of pathlib.Path canditates matching {name}{ext}"""
    if isinstance(path, str):
        path = Path(path)
    path = path  # .resolve()
    files = []
    for ext in extensions:
        pattern = str(path / f"{name}{ext}")
        _files = glob.glob(pattern)
        if _files:
            if break_on_first:
                files = _files
                break
            files += _files
    if not files:
        files = [None]
    else:
        files = [Path(f) for f in files]
    if return_first:
        return files[0]
    return files


def get_junit_files(rootdir):
    junit_file = rootdir / "mystran-testing.xml"
    junit_html_target = junit_file.parent / (junit_file.stem + ".html")
    return junit_file, junit_html_target


def setup(rootdir):
    # -------------------------------------------------------------------------
    # ensure conftest.py and __init__ is there
    _init = rootdir / "__init__.py"
    if not _init.exists():
        _init.touch()
    _conftest = rootdir / "conftest.py"
    if not _conftest.exists():
        with open(_conftest, "w") as fh:
            fh.write("from mystran_validation.conftest_ref import *\n")
    # delete junit stuff
    junit_file, junit_html_target = get_junit_files(rootdir)
    try:
        os.remove(junit_file)
    except FileNotFoundError:
        pass
    try:
        os.remove(junit_html_target)
    except FileNotFoundError:
        pass
    try:
        shutil.rmtree(rootdir / "dumps")
    except FileNotFoundError:
        pass


def teardown(rootdir):
    """clean rootdir"""
    to_delete = ["**/bandit.*", "conftest.py", "__init__.py", "__pycache__"]
    to_delete = [rootdir / p for p in to_delete]
    for pattern in to_delete:
        files = glob.glob(str(pattern), recursive=True)
        logging.debug(f"deleting temporary files:")
        for file in files:
            logging.debug(f" * {file}")
            try:
                shutil.rmtree(file)
            except NotADirectoryError:
                os.remove(file)


def _init_rootdir(rootdir):
    path = rootdir / "example"
    click.echo(click.style(f"creating missing test-cases repository...", bold=True))
    click.echo(click.style(f"created repository {path}", fg="green"))
    path.mkdir(parents=True, exist_ok=True)
    # copy example files
    _files = ["bulk_model.nas", "bulk_model_2.dat", "test_bar.ini", "test_case_03.op2"]
    for f in _files:
        src = Path(pkg_resources.resource_filename("mystran_validation.data", f))
        shutil.copy(src, path / f)
    return path


def _ensure_paths(profile_name, rootdir, mystran_bin):
    # -------------------------------------------------------------------------
    # get configuration
    try:
        config_fpath, config = get_conf()
    except FileNotFoundError:  # if not existing, create with defaults
        click.echo(click.style("Creating missing configuration file...", bold=True))
        config_fpath = init_config(
            profile_name=profile_name, rootdir=rootdir, mystran_bin=mystran_bin
        )
        click.echo(
            click.style(f"created configuration file {config_fpath}", fg="green")
        )
        click.echo("consider to modify this file as per your preferences")
        click.echo("\n")
        config_fpath, config = get_conf()
    # -------------------------------------------------------------------------
    # get profile
    try:
        profile = get_profile(config, profile_name)
    except KeyError:
        click.echo(
            click.style(
                f"profile [{profile_name}] not found in {config_fpath}",
                fg="red",
            )
        )
        default = get_profile(config, "DEFAULT")
        config[profile_name] = {
            "mystran-bin": mystran_bin if mystran_bin else default["mystran-bin"],
            "rootdir": rootdir if rootdir else default["rootdir"],
        }
        with open(config_fpath, "w") as fh:
            config.write(fh)
        click.echo(
            click.style(
                f"creating profile [{profile_name}] with following options:", fg="green"
            )
        )
        profile = get_profile(config, profile_name)
        for k, v in profile.items():
            click.echo(f"  * {k}: {v}")
        click.echo("\n")
    # rootdir may not exist at this point
    rootdir = Path(config[profile_name]["rootdir"])
    if not rootdir.exists():
        _init_rootdir(rootdir)


@click.group()
@click.option(
    "-p", "--profile", default="DEFAULT", type=str, help="configuration title"
)
@click.option("-r", "--rootdir")
@click.option("-m", "--mystran-bin")
@click.option("-l", "--loglevel", default="info", type=str)
@click.pass_context
def main(ctx, profile, rootdir, mystran_bin, loglevel):
    profile_name = profile  # profile will be used for dict
    # -------------------------------------------------------------------------
    # handling logging verbosity
    getattr(logging, loglevel.upper())
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: %s" % loglevel)
    logging.basicConfig(level=numeric_level)
    ctx.ensure_object(dict)
    ctx.obj["rootdir"] = rootdir
    ctx.obj["mystran_bin"] = mystran_bin
    ctx.obj["profile"] = profile


@main.command()
@click.pass_context
def init(ctx):
    # config_fpath, config = get_conf()
    # profile = get_profile(config, profile_name)
    _ensure_paths(
        profile_name=ctx.obj["profile"],
        rootdir=ctx.obj["rootdir"],
        mystran_bin=ctx.obj["mystran_bin"],
    )


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option("--report/--no-report", default=None)
@click.option("--open-report/--not-open-report", default=None)
@click.option("--dump", type=click.IntRange(0, 3))
@click.argument("pytest_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def run(ctx, report, open_report, dump, pytest_args):
    try:
        config_fpath, config = get_conf()
    except FileNotFoundError:
        click.echo(click.style("Configuration file not found!", "red"))
        click.echo(click.style("consider running `mystran-val init` first"))
        sys.exit(1)
    profile = get_profile(config, ctx.obj["profile"])
    # -------------------------------------------------------------------------
    # override with CLI arguments
    mystran_bin = Path(
        ctx.obj["mystran_bin"] if ctx.obj["mystran_bin"] else profile["mystran-bin"]
    )
    rootdir = Path(ctx.obj["rootdir"] if ctx.obj["rootdir"] else profile["rootdir"])
    # -------------------------------------------------------------------------
    # check that mystran binary exists
    if not mystran_bin.exists():
        click.echo(click.style(f"Mystran Binary `{mystran_bin}` not found!", "red"))
        sys.exit(1)
    if not rootdir.exists():
        click.echo(click.style(f"Rootdir `{rootdir}` not found!", "red"))
        sys.exit(1)
    # -------------------------------------------------------------------------
    os.environ["MYSTRAN_BIN"] = str(mystran_bin)
    if dump is None:
        dump = profile["dump"]
        logging.info(f"dumping got from profile {dump=}")
    if report is None:
        # use profile report
        report = bool(int(profile["report"]))
    if open_report is None:
        open_report = bool(int(profile["open-report"]))
    os.environ["MYSTRAN_VALIDATION_DUMP_XLSX"] = str(dump)
    # -------------------------------------------------------------------------
    # setting up rootdir
    setup(rootdir)
    # =========================================================================
    # summary
    # =========================================================================
    click.echo(click.style(f"test-cases params:", fg="green"))
    click.echo(click.style(f"  * mystran_bin={mystran_bin}", fg="green"))
    click.echo(click.style(f"  * rootdir={rootdir}", fg="green"))
    click.echo(click.style(f"  * {dump=}", fg="green"))
    click.echo("\n")
    args = list(pytest_args)
    if report:
        junit_file, junit_html_target = get_junit_files(rootdir)
        args += [f"--junitxml={junit_file}"]
    args.append(str(rootdir))
    pytest.main(args)
    if report:
        python_bin = shutil.which("python")
        subprocess.run(
            shlex.split(
                f"{python_bin} -m junit2htmlreport {junit_file} {junit_html_target}"
            )
        )
    teardown(rootdir)
    if report and open_report:
        webbrowser.open(str(junit_html_target.resolve()))
    return 0


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("pytest_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def collect(ctx, pytest_args):
    try:
        config_fpath, config = get_conf()
    except FileNotFoundError:
        click.echo(click.style("Configuration file not found!", "red"))
        click.echo(click.style("consider running `mystran-val init` first"))
        sys.exit(1)
    profile = get_profile(config, ctx.obj["profile"])
    rootdir = Path(ctx.obj["rootdir"] if ctx.obj["rootdir"] else profile["rootdir"])
    setup(rootdir)
    args = list(pytest_args) + ["--collect-only"]
    args.append(str(rootdir))
    pytest.main(args)
    teardown(rootdir)
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-

"""Top-level package for MYSTRAN validation."""

__author__ = """Nicolas Cordier"""
__email__ = "nicolas.cordier@numeric-gmbh.ch"
__version__ = "0.8.0"

import os
import configparser
from pathlib import Path
import logging
import shutil
from functools import wraps
import asyncio
import datetime as dt
from appdirs import AppDirs


def assert_frame_equal(dfa, dfb, atol, rtol):
    """it's hard to understand how pandas tesing is actually working
    better to fall back to numpy testing, assuming we do not need to
    check columns, index, etc...
    """
    diff = abs(dfa - dfb)
    crit = atol + rtol * abs(dfb)
    failing = diff[diff > crit]
    # keep rows and columns where at least one failure occurs
    failures = failing.dropna(how="all").dropna(how="all", axis=1)
    abs_error = failing.max().max()
    rel_error = (failing / abs(dfb)).max().max()
    failing = diff > crit
    return failing, failures, abs_error, rel_error


def background(func):
    """decorator for background jobs"""

    @wraps(func)
    def wrapped(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(None, func, *args, *kwargs)

    return wrapped


# =============================================================================
# CONFIGURATION stuff and default options
# =============================================================================
DIRS = AppDirs("mystran-validation", "numeric")

DEFAULTS = {
    "DEFAULT": {
        "mystran-bin": os.getenv("MYSTRAN_BIN", shutil.which("mystran")),
        "rootdir": os.path.join(DIRS.user_data_dir, "mystran-test-cases"),
        "# dump": "0: never, 1: failures only, 2: always",
        "dump": "1",
        "report": "1",
        "open-report": "1",
    }
}


def init_config(profile_name="DEFAULT", **kwargs):
    """create/update a profile in the config file. If config file not found,
    create it from scratch
    """
    config_fpath = Path(DIRS.user_config_dir) / "config.ini"
    parser = configparser.ConfigParser()
    params = DEFAULTS.copy()
    if profile_name not in params:
        params[profile_name] = {}
    for k, v in kwargs.items():
        if not v:
            continue
        params[profile_name][k.replace("_", "-")] = v
    parser.read_dict(params)
    config_fpath.parent.mkdir(parents=True, exist_ok=True)
    with open(config_fpath, "w") as configfile:
        parser.write(configfile)
    logging.debug(f"created {config_fpath}")
    return config_fpath


def is_comment(param):
    return param.strip()[0] in ("#", ";")


def get_conf():
    config_fpath = Path(DIRS.user_config_dir) / "config.ini"
    if not config_fpath.exists():
        raise FileNotFoundError(f"Configuration file '{str(config_fpath)}' not found")
        # init_config()
    parser = configparser.ConfigParser()
    parser.read(config_fpath)
    # -------------------------------------------------------------------------
    # always check if "DEFAULT" is up-to-date
    defaults = set((p for p in DEFAULTS["DEFAULT"].keys() if not is_comment(p)))
    actual = set((p for p in parser["DEFAULT"].keys() if not is_comment(p)))
    if defaults ^ actual:
        # some changes will occur, backup initial config
        backup_fpath = config_fpath.parent / (
            config_fpath.stem
            + ".ini.backup_"
            + dt.datetime.now().isoformat(timespec="minutes")
        )
        with open(backup_fpath, "w") as configfile:
            parser.write(configfile)
        logging.info(f"backuped {backup_fpath} before modification")
    # missing parameters
    for missing_param in defaults - actual:
        parser["DEFAULT"][missing_param] = DEFAULTS["DEFAULT"][missing_param]
        logging.info(f"added missing parameter {missing_param}")
    # unknown parameters
    for unknown_param in actual - defaults:
        _val = parser["DEFAULT"].pop(unknown_param)
        logging.warning(f"removed unknown parameter {unknown_param}={_val}")
    with open(config_fpath, "w") as configfile:
        parser.write(configfile)
    return config_fpath, parser


def get_profile(config, profile_name=None):
    if profile_name:
        try:
            config = dict(config[profile_name].items())
        except KeyError:
            msg = f"{profile_name} not defined in configuration file"
            raise KeyError(msg)
    else:
        config = dict(config["DEFAULT"].items())
    for k, v in config.items():
        config[k] = os.path.expanduser(v)
    return config



# content of conftest.py
import ast
import configparser
import numbers
import os
import shlex
import shutil
import subprocess
import traceback
from pathlib import Path

import pandas as pd
import pytest
from femap_neutral_parser import Parser as NeuParser

from mystran_validation import assert_frame_equal, background
from mystran_validation.parsers import subset
from mystran_validation.parsers.nastran_op2 import Parser as OP2Parser

# =============================================================================
# test collection
# =============================================================================
ROOTDIR = None


# def pytest_collection_modifyitems(items):
#     """add CARDs to marks"""
#     for item in items:
#         bulk_file = item.spec["bulk"]
#         with open(bulk_file, "r") as fh:
#             lines = fh.readlines()
#         words = set(
#             [
#                 l.split(" ")[0].strip()
#                 for l in lines
#                 if not l.startswith("$") and not l.startswith("+")
#             ]
#         ) - set(["", " "])
#         for tag in words:
#             item.add_marker(getattr(pytest.mark, tag))


def pytest_collect_file(parent, path):
    global ROOTDIR
    if ROOTDIR is None:  # set rootdir for centralized dumping
        ROOTDIR = Path(path.dirpath())
    _path = Path(path)
    if _path.suffix == ".ini" and _path.name.startswith("test"):
        return IniTestFile.from_parent(parent, fspath=path)


class IniTestFile(pytest.File):
    bulkfile = None
    rootdir = None
    rootname = None
    mystran_run_status = {}
    reference = {}
    op2s = {}

    def collect(self):
        config = configparser.ConfigParser()
        config.read(self.fspath)
        # ---------------------------------------------------------------------
        # prepare paths
        self.rootdir = Path(self.fspath).parent
        self.rootname = Path(self.fspath).stem
        # ---------------------------------------------------------------------
        # clean working dir
        wdir = self.rootdir / (".out_" + self.rootname)
        try:
            wdir.mkdir()
        except FileExistsError:
            shutil.rmtree(wdir)
            wdir.mkdir()
        for name in config.sections():
            # =================================================================
            # building and yielding atomic test [name]
            # =================================================================
            spec = dict(config[name].items())
            bulkfile = self.rootdir / spec["bulk"]
            if self.bulkfile:
                if bulkfile != self.bulkfile:
                    raise ValueError(f"{name}: One class, one bulk!")
            spec["workingdir"] = wdir
            spec["bulk"] = bulkfile
            ref = spec["reference"]
            # -----------------------------------------------------------------
            # if ref is obviously a file
            if Path(ref).suffix.lower() in (".op2", ".neu"):
                spec["reference"] = self.rootdir / spec["reference"]
            # -----------------------------------------------------------------
            # if ref a sing value
            else:
                ref = float(ref)
                spec["reference"] = ref
            spec["output"] = spec["workingdir"] / (bulkfile.stem + ".NEU")
            spec["rtol"] = float(spec.get("rtol", 1e-05))
            spec["atol"] = float(spec.get("atol", 1e-08))
            yield IniItem.from_parent(self, name=name, spec=spec)


class IniItem(pytest.Item):
    """Atomic test. This is where one single [test] is performed"""

    def __init__(self, name, parent, spec):
        super().__init__(name, parent)
        self.spec = spec

    def run_mystran(self):
        """run bulkfile"""
        target = shutil.copy(self.spec["bulk"], self.spec["workingdir"])
        cmd = f"{os.getenv('MYSTRAN_BIN')} {target}"
        status = subprocess.run(shlex.split(cmd))
        # ---------------------------------------------------------------------
        # get actual results
        neu = NeuParser(self.spec["output"], autotranslate=False)
        self.parent.actual = neu
        self.parent.actual.info(doprint=False)  # pre-digest data
        self.parent.actual_available_vectors = sorted(
            self.parent.actual.output_vectors.keys()
        )
        return status

    def expected(self, vector, filters=None, axis=None, raw=False):
        """build reference (expected) DataFrame"""
        if not filters:
            filters = {}
        df_expected = None
        # ---------------------------------------------------------------------
        # reference is a numerical value
        if isinstance(self.spec["reference"], numbers.Number):
            if axis is None:
                raise ValueError("param `axis` must be specified for manual checking")
            _data = filters.copy()
            _data.update({axis: [self.spec["reference"]]})
            df_expected = pd.DataFrame(_data)
            df_expected.set_index(list(filters.keys()), inplace=True)
            df_expected = df_expected[axis].to_frame()
        # ---------------------------------------------------------------------
        # reference is a ref file
        elif self.spec["reference"].suffix.lower() == ".op2":
            op2 = self.parent.op2s[self.spec["reference"]]
            df_expected = op2.get_vector(vector=vector, raw=raw, **filters)
        else:
            raise ValueError("reference {self.spec['reference']} not understood")
        return df_expected

    def actual(self, vector, filters=None, axis=None, raw=False):
        """build actual DataFrame"""
        if not filters:
            filters = {}
        # -----------------------------------------------------------------
        # get values from neutral file, and reshape it
        df_actual = pd.DataFrame(getattr(self.parent.actual, f"get_{vector}")(raw=raw))
        if not raw:
            if "NodeID" in df_actual and "ElementID" not in df_actual:
                df_actual.set_index(["SubcaseID", "NodeID"], inplace=True)
            elif "ElementID" in df_actual and "NodeID" not in df_actual:
                df_actual.set_index(["SubcaseID", "ElementID"], inplace=True)
            else:
                raise ValueError("cannot get values")
            df_actual.sort_index(inplace=True)
            df_actual = subset(df_actual, **filters)
            if axis:
                try:
                    df_actual = df_actual[axis].to_frame()
                except KeyError:
                    msg = f"{axis} is not a proper column. Use one of {df_actual.columns.tolist()}"
                    raise KeyError(msg)
        return df_actual

    def runtest(self):
        """trigger pytest"""
        if self.spec.get("skip"):
            pytest.skip(self.spec["skip"])
        # ---------------------------------------------------------------------
        # run Mystran only once
        if self.spec["bulk"] not in self.parent.mystran_run_status:
            # first test for this class, a few things to set up
            self.parent.mystran_run_status[self.spec["bulk"]] = self.run_mystran()
        # process OP2 only once
        ref = self.spec["reference"]
        if not isinstance(ref, numbers.Number):
            if ref.suffix.lower() == ".op2" and str(ref) not in self.parent.op2s:
                self.parent.op2s[ref] = OP2Parser(str(self.spec["reference"]))
            self.op2 = self.parent.op2s[ref]  # shortcut for self
        else:
            self.op2 = None
        self.user_properties += [("ref", self.op2)]
        self.user_properties += [("description", self.spec["description"])]
        self.user_properties += [("atol", self.spec["atol"])]
        self.user_properties += [("rtol", self.spec["rtol"])]

        try:
            vector = self.spec["vector"]
            # ---------------------------------------------------------------------
            # get subset index
            filters = dict(
                SubcaseID=ast.literal_eval(self.spec.get("subcaseids", "None")),
                NodeID=ast.literal_eval(self.spec.get("nodeids", "None")),
                ElementID=ast.literal_eval(self.spec.get("elementids", "None")),
            )
            axis = self.spec.get("axis")
            filters = {k: [v] for k, v in filters.items() if v}
            # ---------------------------------------------------------------------
            # get dataframes to compare
            df_actual = self.actual(vector, filters, axis)
            df_expected = self.expected(vector, filters, axis)
            # -----------------------------------------------------------------
            # check tolerances
            rtol = self.spec["rtol"]
            atol = self.spec["atol"]
            failing, failures, aerr, rerr = assert_frame_equal(
                df_actual, df_expected, rtol=rtol, atol=atol
            )
        except Exception as exc:
            tb = traceback.format_exc()
            raise GenericException(self, exc, tb)
        else:
            if len(failures) > 0:
                raise IniException(
                    self, df_actual, df_expected, failures, rtol, atol, aerr, rerr
                )
        finally:
            # dump comparisons performed
            if (
                os.getenv("MYSTRAN_VALIDATION_DUMP_XLSX") == "2"
                or os.getenv("MYSTRAN_VALIDATION_DUMP_XLSX") == "1"
                and len(failures) > 0
            ):
                try:
                    dump(
                        df_expected,
                        df_actual,
                        wdir=self.spec["workingdir"],
                        sheetname=self.spec["vector"],
                        name=Path(self.parent.name),
                        failing=failing,
                    )
                except Exception as exc:
                    dump(
                        df_expected,
                        df_actual,
                        wdir=self.spec["workingdir"],
                        sheetname=self.spec["vector"],
                        name=Path(self.parent.name),
                        failing=failing,
                        debug=True,
                    )

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        if isinstance(excinfo.value, IniException):
            (
                item,
                df_actual,
                df_expected,
                failures,
                rtol,
                atol,
                aerr,
                rerr,
            ) = excinfo.value.args
            df_actual = df_actual.loc[failures.index]
            df_expected = df_expected.loc[failures.index]
            # failing difference
            return "\n".join(
                [
                    f"usecase `{self.fspath}::[{self.name}]`\nexecution failed given precision requirements:\n  * {atol=}\n  * {rtol=}\n",
                    f"failing with:\n  * Absolute difference {aerr=}\n  * Relative difference {rerr=}\n",
                    f"Expected\n--------\n{df_expected}\n",
                    f"Actual\n------\n{df_actual}",
                ]
            )
        elif isinstance(excinfo.value, GenericException):
            item, exc, traceback = excinfo.value.args
            return "\n".join(
                [
                    f"usecase `{self.fspath}::[{self.name}]` raised the following exception\n",
                    f"{exc}\n",
                    f"file: {traceback}",  # get rid of "file " prefix
                ]
            )

    def reportinfo(self):
        return self.fspath, 0, f"usecase: {self.name}"


class IniException(Exception):
    pass


class GenericException(Exception):
    pass


def apply_color(x, failing):
    colors = {False: "green", True: "red; font-weight: bold"}
    return failing.applymap(lambda val: "color: {}".format(colors.get(val, "")))


def highlight_columns(x):
    if x.name[0] == "Actual":
        style = f"background-color: azure"
    else:
        style = f"background-color: beige"
    return [style] * x.shape[0]


# @background
def dump(df_expected, df_actual, wdir, name, sheetname, failing, debug=False):
    full = (
        pd.concat({"Expected": df_expected, "Actual": df_actual})
        .unstack(level=0)
        .swaplevel(axis=1)
        .sort_index(axis=1)
    )
    # reshape failing to have same format as `full`
    failing = (
        pd.concat({"Expected": failing, "Actual": failing})
        .unstack(level=0)
        .swaplevel(axis=1)
        .sort_index(axis=1)
    )
    # =========================================================================
    # formatting
    # =========================================================================
    dumping_dir = ROOTDIR / "dumps"
    dumping_dir.mkdir(exist_ok=True)
    target = dumping_dir / (name.stem + ".xlsx")
    mode = "a" if target.exists() else "w"
    # style dataframe
    styled = full.style.apply(apply_color, axis=None, failing=failing)
    styled = styled.apply(highlight_columns)
    with pd.ExcelWriter(target, mode=mode) as writer:
        styled.to_excel(writer, sheet_name=sheetname)
    # formatting
    # if len(failures) == 0:
    #     return
    # cols = df_expected.columns.tolist()
    # guilty_columns
    # breakpoint()
    # writer = pd.ExcelWriter(target, mode="a")
    # workbook  = writer.book
    # worksheet = writer.sheets[sheetname]
    # red_format = workbook.add_format({'bg_color':'red'})
    # green_format = workbook.add_format({'bg_color':'green'})

    # worksheet.conditional_format('B2:B4', {'type': 'text',
    #                                       'criteria': 'containing',
    #                                        'value':     'Fail',
    #                                        'format': red_format})

    # worksheet.conditional_format('B2:B4', {'type': 'text',
    #                                       'criteria': 'containing',
    #                                        'value':   'Pass',
    #                                        'format':  green_format})
    # writer.save()
